from typing import Dict

REPORT_VARIABLES = [
    '[[protocol_name_bold]]',
    '[[protocol_repository_bold]]',
    '[[protocol_commit_bold]]', 
    '[[protocol_date_bold]]',
    '[[type_bold]]',
    '[[protocol_sloc_bold]]',
]

INFORMATION_TABLE_VARIABLES: Dict[str, str] = {
    '[[protocol_protocol_name_bold]]': 'Protocol Name',
    '[[protocol_repository_bold]]': 'Repository',
    '[[protocol_commit_bold]]': 'Commit', 
    '[[protocol_date_bold]]': 'Date',
    '[[protocol_type_bold]]': 'Protocol Type',
    '[[protocol_sloc_bold]]': 'SLOC'
}

REPORT_SECTION_HEADINGS = [
    "1. About", "2. Disclaimer",
    "3. Introduction", "4. Risk Classification",
    "4.1. Impact", "4.2. Likelihood",
    "4.3. Action required for severity levels",
    "5. Executive Summary", "6. Findings"
]

PROTOCOL_INFORMATION_TABLE_COLUMNS = ["Protocol Name", "Repository", "Commit", "Date", "Protocol Type", "SLOC"]

SEVERITY_CLASSIFICATION_TABLE_ROWS = [
    ["Likelihood: High", "Critical", "High", "Medium"],
    ["Likelihood: Medium", "High", "Medium", "Low"],
    ["Likelihood: Low", "Medium", "Low", "Low"],
]
SEVERITY_CLASSIFICATION_TABLE_COLUMNS = ["Severity", "Impact: High", "Impact: Medium", "Impact: Low"]

FINDING_COUNT_TABLE_COLUMNS = ["Severity", "Amount"]


from urllib.parse import urlparse


class Utils:
    """
        Miscellaneous utility functions.
    """
    
    @staticmethod
    def validate_url(url: str) -> bool:
        result = urlparse(url)
        return all([result.scheme, result.netloc])


ALREADY_INITIALIZED_MSG: str = "The current directory is not initialized. Please run [bold]\"compiloor init\"[/bold] to initialize the current directory."
NOT_INITIALIZED_MSG: str = "The current directory is already initialized. If you want to override the current directory, run [bold]\"compiloor init --force\"[/bold]."



from compiloor.services.logger.logger import Logger
from compiloor.services.utils.utils import Utils


class ConfigUtils:
    """
        Miscellaneous functions for the configuration file.
    """
    
    @staticmethod
    def validate_config_template_urls(config: dict) -> None:
        # Checking whether the keys are present:
        template_url_is_valid = "template_url" in config         
        stylesheet_url_is_valid = "stylesheet_url" in config
        
        # Checking whether the URLs are valid:
        if stylesheet_url_is_valid: stylesheet_url_is_valid = Utils.validate_url(config["stylesheet_url"])
        if template_url_is_valid: template_url_is_valid = Utils.validate_url(config["template_url"])

        if not template_url_is_valid and not stylesheet_url_is_valid:
            Logger.error("The template URL and the stylesheet URL must be valid URLs.")
            exit(1)
            
        if not template_url_is_valid:
            Logger.error("The template URL must be a valid URL.")
            exit(1)
            
        if not stylesheet_url_is_valid:
            Logger.error("The stylesheet URL must be a valid URL.")
            exit(1)


from fitz import (
    Document, Page, Rect, TEXT_ALIGN_CENTER,
    TEXT_ALIGN_RIGHT, PDF_REDACT_IMAGE_NONE, PDF_ENCRYPT_KEEP
)

from compiloor.services.typings.utils import PAGE_NUMBER_FONT_SIZE, PRIMARY_COLOR_IN_PERCENTAGES


class PageNumberingUtils:
    """
        A class containing methods for adding page numbers and finding legend to a report.
    """

    # TODO: Rewrite this method with more elegant code.    
    @staticmethod
    def create_report_with_page_numbers_and_legend(report_path: str, report_section_headings: list[str]) -> None:
        """
            Creates a new report with page numbers and a legend for the findings.
        """
    
        new_pdf = Document(report_path) # Using fitz to edit the PDF.
            
        pages_to_delete: list[int] = []
        
        page: Page    
        
        page_to_section_index: dict[str, int] = {}
        page_to_section_index_occurences: dict[str, int] = {}
        
        for page in new_pdf.pages():
            if page.get_textpage().extractText() == "": 
                pages_to_delete.append(page.number) 
                continue
            page.clean_contents()
            
        if len(pages_to_delete): new_pdf.delete_pages(pages_to_delete) # Deleting empty pages.
        
        # Normalizing the page orientations:    
        for page in new_pdf.pages():
            if page.number == 0: continue         
            x1, y1 = page.cropbox.x1, page.cropbox.y1

            page.clean_contents()
            
            # Insert page numbers on each page:
            page.insert_textbox(
                # The whole width of the page for the last 18th of the page vertically:
                Rect(0, y1 - y1 / 18, x1, y1),
                str(page.number),
                align = TEXT_ALIGN_CENTER,
                fontsize = PAGE_NUMBER_FONT_SIZE,
                color = PRIMARY_COLOR_IN_PERCENTAGES
            )
            
            for section_index in range(len(report_section_headings)):
                heading = report_section_headings[section_index]
                
                if not heading in page.get_textpage().extractText(): continue
                
                if heading in page_to_section_index: continue
                
                if not heading in page_to_section_index_occurences: page_to_section_index_occurences[heading] = 0
                
                page_to_section_index_occurences[heading] += 1
                
                if page_to_section_index_occurences[heading] < 2: continue

                page_to_section_index[heading] = page.number

        for section in page_to_section_index.keys():
            for page in new_pdf.pages():
                section_index = section.split(" ")[0]

                section_index = section_index if not "." in section_index else section_index[:-1]
                
                search_fragment: str = '{{[' + section_index + ']_page}}' if not '[' in section_index and not ']' in section_index else '{{' + section_index + '_page}}'
                fragment = page.search_for(search_fragment)
                if not fragment: continue

                page.add_redact_annot(
                    fragment[0],
                    str(page_to_section_index[section]),
                    cross_out=False,
                    align=TEXT_ALIGN_RIGHT,
                    fontsize = PAGE_NUMBER_FONT_SIZE,
                    text_color = PRIMARY_COLOR_IN_PERCENTAGES,
                    fontname="Times-Roman"
                )
                page.apply_redactions(images=PDF_REDACT_IMAGE_NONE)
                break
            
        new_pdf.save(report_path, incremental=True, encryption=PDF_ENCRYPT_KEEP)


from typer import Typer

from typing_extensions import Annotated

from compiloor.services.environment.utils import FileUtils
from compiloor.services.parser.chromium import ChromiumUtils
from compiloor.services.parser.page_numbering import PageNumberingUtils
from compiloor.services.parser.utils import ReportCustomizer
from compiloor.services.typings.finding import SeverityAnnotation
from compiloor.services.environment import (
    current_directory_initialized, initialize_directory,
    add_finding_template
)
from compiloor.services.environment.constants import INITIALIZED, NOT_INITIALIZED
from compiloor.services.utils.config import ConfigUtils

cli = Typer()

@cli.command("init")
def init(force: Annotated[bool, "force"] = False):
    current_directory_initialized(NOT_INITIALIZED, force)
    initialize_directory(force)
    
@cli.command("add-finding")
def add_finding(severity: Annotated[SeverityAnnotation, "severity"] = SeverityAnnotation.MEDIUM):
    current_directory_initialized(INITIALIZED)
    add_finding_template(severity.cast_to_severity())
    
    
@cli.command("compile")
def compile_report():
    current_directory_initialized(INITIALIZED)
    ConfigUtils.validate_config_template_urls(FileUtils.read_config(json=True))
    
    # The customizer creates the base report and handles almost all serialization:
    customizer = ReportCustomizer()
    
    # Save the initial report to a PDF file:
    (_, report_path) = ChromiumUtils.create_chromium_document(customizer.report)
    
    # Finalize the report by adding page numbers and a legend:
    PageNumberingUtils.create_report_with_page_numbers_and_legend(report_path, customizer.report_section_headings)



from rich import print as rich_print

class Logger:
    
    @staticmethod
    def error(message: str = "", *args, **kwargs) -> None:
        rich_print(f"[bold red]Error: [/bold red][white]{message}", *args, **kwargs)

    @staticmethod
    def warning(message: str = "", *args, **kwargs) -> None:
        rich_print(f"[bold yellow]Warning: [/bold yellow][white]{message}", *args, **kwargs)
        
    @staticmethod
    def QA(message: str = "", *args, **kwargs) -> None:
        rich_print(f"[bold blue]QA: [/bold blue][white]{message}", *args, **kwargs)
        
    @staticmethod
    def success(message: str = "", *args, **kwargs) -> None:
        rich_print(f"[bold green]Success: [/bold green][white]{message}", *args, **kwargs)


from os.path import abspath


FINDING_TEMPLATE_NAME = "finding-template.md"
CONFIG_NAME = "config.json"

INITIALIZED = True
NOT_INITIALIZED = False

MAIN_DIRECTORY = "compiloor-report"
MAIN_DIRECTORY_ABS_PATH = abspath(MAIN_DIRECTORY)

FINDINGS_DIRECTORY = "findings"
REPORTS_DIRECTORY = MAIN_DIRECTORY + "/reports"
TEMPLATE_DIRECTORY = "templates"

CONFIG_DIRECTORY = MAIN_DIRECTORY + "/" + CONFIG_NAME

BASE_CONFIG_SCHEMA: dict[str, str] = {
    "title": "SECURITY RESEARCHER NAME",
    "author": "SECURITY RESEARCER",
    "date": "DATE",
    "company_name": "COMPANY NAME",
    "protocol_name": "PROTOCOL NAME",
    "repository": "GITHUB/GITLAB REPOSITORY",
    "type": "SOME PROTOCOL TYPE",
    "commit": "COMMIT HASH",
    "sloc": "SLOC",
    "about_content": "ABOUT SECTION",
    "disclaimer_content": "DISCLAIMER CONTENT",
    "introduction_content": "INTRODUCTION CONTENT",
    "template_url": "REPORT URL",
    "stylesheet_url": "REPORT CONTENT",
    "cover_img_url": "COVER IMAGE URL",
}

BASE_FINDING_MD_TEMPLATE: str = """# [{{finding_severity}}-{{finding_index}}]

## Severity

**Impact:**

**Likelihood:**

## Description

## Recommendations
"""


from json import dumps

from os import mkdir
from os.path import abspath, isdir

from shutil import rmtree

from compiloor.services.environment.constants import (
    BASE_CONFIG_SCHEMA, BASE_FINDING_MD_TEMPLATE,
    FINDINGS_DIRECTORY, MAIN_DIRECTORY, REPORTS_DIRECTORY
)
from compiloor.services.environment.utils import FileUtils, FindingUtils
from compiloor.services.typings.finding import Severity
from compiloor.services.logger.messages import ALREADY_INITIALIZED_MSG, NOT_INITIALIZED_MSG
from compiloor.services.logger import Logger


# TODO: Rewrite the following functions into a utility class:
def current_directory_initialized(initialized: bool = True, force: bool = False) -> None:
    if force or _current_directory_initialized() == initialized: return
    Logger.error(ALREADY_INITIALIZED_MSG if initialized else NOT_INITIALIZED_MSG)
    return exit(1)



def _current_directory_initialized() -> bool:
    return isdir(abspath(MAIN_DIRECTORY)) and isdir(abspath(FINDINGS_DIRECTORY))

def initialize_directory(force: bool = False) -> None:
    # erase everything if force:
    
    _abs_main_dir = abspath(MAIN_DIRECTORY)
    _abs_findings_dir = abspath(FINDINGS_DIRECTORY)
    
    if force: rmtree(_abs_main_dir, ignore_errors=True)

    try:
        if not isdir(_abs_findings_dir): mkdir(_abs_findings_dir)

        mkdir(_abs_main_dir)
        mkdir(abspath(REPORTS_DIRECTORY))
        
        open(f"{_abs_main_dir}/config.json", "w").write(dumps(BASE_CONFIG_SCHEMA, indent=4))

        Logger.success("Successfully initialized the directory!")
        
    except Exception as err:
        Logger.error(f"An error occured while initializing the directory: {err}")
        exit(1)

def add_finding_template(severity: Severity) -> None:
    try:
        finding_template = BASE_FINDING_MD_TEMPLATE

        folder_sig = severity.cast_to_folder_sig().value
        
        index = FileUtils.get_fs_sig_index(FindingUtils.get_current_finding_amount(severity) + 1)

        # setting the index and severity:
        finding_template = finding_template.replace("{{finding_severity}}", folder_sig)
        finding_template = finding_template.replace("{{finding_index}}", index)    
        
        open(f"{abspath(FINDINGS_DIRECTORY)}/[{folder_sig}-{index}].md", "w").write(finding_template)
        
        Logger.success(f'Successfully added a finding template for severity "{severity.name}"!')
    except Exception as err:
        Logger.error(f"An error occured while adding a finding template: {err}")
        exit(1)


from enum import Enum

FINDING_LIST_TABLE_COLUMNS = ["ID", "Title", "Severity"]

class Severity(Enum):
    """
        Enum class for severity levels.
    """
    
    GAS = 1
    QA = 2
    LOW = 3
    MEDIUM = 4
    HIGH = 5
    CRITICAL = 6
    
    def __str__(self) -> str:
        return str(self.value)
    
    def __repr__(self) -> str:
        return str(self)
    
    def cast_to_folder_sig(self: "Severity") -> "SeverityFolderIndex":
        mapping = {
            Severity.GAS: SeverityFolderIndex.GAS,
            Severity.QA: SeverityFolderIndex.QA,
            Severity.LOW: SeverityFolderIndex.LOW,
            Severity.MEDIUM: SeverityFolderIndex.MEDIUM,
            Severity.HIGH: SeverityFolderIndex.HIGH,
            Severity.CRITICAL: SeverityFolderIndex.CRITICAL,
        }
        return mapping[self]
    
    def cast_to_display_case(self: "Severity") -> str:
        mapping = {
            Severity.GAS: "GAS",
            Severity.QA: "QA",
            Severity.LOW: "Low",
            Severity.MEDIUM: "Medium",
            Severity.HIGH: "High",
            Severity.CRITICAL: "Critical",
        }
        return mapping[self]
    
class SeverityAnnotation(Enum):
    """Annotation class for severity levels."""
    GAS = "gas"
    QA = "QA"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    CRITICAL = "critical"
    
    def cast_to_severity(self: "SeverityAnnotation") -> Severity:
        mapping = {
            SeverityAnnotation.GAS: Severity.GAS,
            SeverityAnnotation.QA: Severity.QA,
            SeverityAnnotation.LOW: Severity.LOW,
            SeverityAnnotation.MEDIUM: Severity.MEDIUM,
            SeverityAnnotation.HIGH: Severity.HIGH,
            SeverityAnnotation.CRITICAL: Severity.CRITICAL,
        }
        return mapping[self]
    
class SeverityFolderIndex(Enum):
    """Enum class for severity levels."""
    GAS = "GAS"
    QA = "QA"
    LOW = "L"
    MEDIUM = "M"
    HIGH = "H"
    CRITICAL = "C"
    
    def cast_to_severity(self: "SeverityFolderIndex") -> Severity:
        mapping = {
            SeverityFolderIndex.GAS: Severity.GAS,
            SeverityFolderIndex.QA: Severity.QA,
            SeverityFolderIndex.LOW: Severity.LOW,
            SeverityFolderIndex.MEDIUM: Severity.MEDIUM,
            SeverityFolderIndex.HIGH: Severity.HIGH,
            SeverityFolderIndex.CRITICAL: Severity.CRITICAL,
        }
        return mapping[self]


from typing import Callable, Dict

from tabulate import tabulate

from compiloor.services.typings.finding import FINDING_LIST_TABLE_COLUMNS, Severity
from compiloor.services.typings.utils import TABLE_HTML_OUTPUT_FORMAT_SETTING
from compiloor.services.typings.config import ProtocolInformationConfigDict
from compiloor.services.parser.finding import Finding
from compiloor.services.typings.report import (
    FINDING_COUNT_TABLE_COLUMNS, INFORMATION_TABLE_VARIABLES,
    SEVERITY_CLASSIFICATION_TABLE_COLUMNS, SEVERITY_CLASSIFICATION_TABLE_ROWS
)


class HTMLTableUtils:
    """
        A class containing utilities for creating HTML tables.
        
        **Disclaimer**: All of the tables use the '[[variable]]' notation for dynamic values instead of '{{variable}}' on purpose.
    """
    
    @staticmethod
    def create_html_table(
        rows: list[list[str]],
        columns: list[str],
        title: str = "",
        highlight_last_row: bool = False, tag: str = "h3"
    ) -> str:
        """
        Creates a table from the given rows and columns.
        """
        
        table: str = f'''
        <div class="table-wrapper">
            [[title]]
            {tabulate(rows, headers=columns, tablefmt=TABLE_HTML_OUTPUT_FORMAT_SETTING)}
        </div>
        '''
        
        table = table.replace("[[title]]", "" if not title else f"<{tag}>{title}</{tag}>")

        if highlight_last_row:
            row = rows[len(rows) - 1]
            table = table.replace(f'<tr><td>{row}', f'<tr class="bold-text"><td>{row}')

        return str(table)

class CustomTableUtils(HTMLTableUtils):
    """
        A class containing utilities for creating custom tables.
    """
    
    @staticmethod
    def clean_row(text: str) -> str:
        """
            Cleans a row of the '[[_bold]]' fragment.
        """
        
        return text.replace('_bold', '').replace('[[', '').replace(']]', '')[9:]
    
    @staticmethod
    def create_severity_classification_table(name: str = "") -> str:
        """
            Creates a severity classification table.
        
            | ------------------ | ------------ | -------------- | ----------- |\n
            | Severity           | Impact: High | Impact: Medium | Impact: Low |\n
            | -------------------| ------------ | -------------- | ----------- |\n
            | Likelihood: High   | Critical     | High           | Medium      |\n
            | -------------------| ------------ | -------------- | ----------- |\n
            | Likelihood: Medium | High         | Medium         | Low         |\n
            | -------------------| ------------ | -------------- | ----------- |\n
            | Likelihood: Low    | Medium       | Low            | Low         |\n
            | -------------------| ------------ | -------------- | ----------- |\n
        """
        
        return TableUtils.create_html_table(
            SEVERITY_CLASSIFICATION_TABLE_ROWS,
            SEVERITY_CLASSIFICATION_TABLE_COLUMNS,
            name
        )
        
    
    @staticmethod
    def create_information_table(
        config: ProtocolInformationConfigDict,
        name: str  = "Protocol Summary",
        information_row_heading_tag: str = "b"
    ) -> str:
    
        """
            Creates a protocol information table from the given config.
            
            | -------------- | ----------------------- |\n
            | Protocol Name  | Example Name            |\n
            | -------------- | ----------------------- |\n
            | Repository     | Some repo               |\n
            | -------------- | ----------------------- |\n
            | Commit         | 3f5ad4..                |\n
            | -------------- | ----------------------- |\n
            | Date           | September 14th 2023     |\n
            | -------------- | ----------------------- |\n
            | Type           | Lending Protocol        |\n
            | -------------- | ----------------------- |\n
            | SLOC           | 500                     |\n
            | -------------- | ----------------------- |\n

        """
                
        table: str = HTMLTableUtils.create_html_table(
            [
                [
                    row, config[CustomTableUtils.clean_row(row)]
                ] for row in INFORMATION_TABLE_VARIABLES.keys()
            ], 
            [], # PROTOCOL_INFORMATION_TABLE_COLUMNS,
            name
        )
        
        for _key in INFORMATION_TABLE_VARIABLES.keys():
            table = table.replace(
            _key,
            f'<{information_row_heading_tag}>{INFORMATION_TABLE_VARIABLES[_key]}</{information_row_heading_tag}>'
        )
            
        return table
    
    @staticmethod
    def create_findings_count_table(
        finding_amounts_by_severity: dict[Severity, int],
        total_findings_amount: int,
        name: str = "Findings Count"
    ) -> str:
        """
            Creates a findings count table from the given finding counts.
        
            | -------------- | ----------------------- |\n
            | Severity       | Amount                  |\n
            | -------------- | ----------------------- |\n
            | Critical       | 0                       |\n
            | -------------- | ----------------------- |\n
            | High           | 0                       |\n
            | -------------- | ----------------------- |\n
            | Medium         | 0                       |\n
            | -------------- | ----------------------- |\n
            | Low            | 0                       |\n
            | -------------- | ----------------------- |\n
            | QA             | 0                       |\n
            | -------------- | ----------------------- |\n
            | GAS            | 0                       |\n
            | -------------- | ----------------------- |\n
            | Total Findings | 0                       |\n
            | -------------- | ----------------------- |\n
        """
        
        finding_counts = [
            [severity.cast_to_display_case(), finding_amounts_by_severity[severity]]
            for severity in finding_amounts_by_severity
        ]
        
        # Doesn't allow me to do a one-liner because reverse doesn't return the mutated array, but rather mutates the pointer of the array.
        finding_counts.reverse()

        finding_counts += [["Total Findings", total_findings_amount]]     
            
        return TableUtils.create_html_table(
            finding_counts,
            FINDING_COUNT_TABLE_COLUMNS,
            name,
            True
        )
    
    def create_findings_summary_table(findings: list[Finding], name: str = "Summary of Findings") -> str:
        """
            Creates a summary of findings table from the given findings.
        
            | -------------- | ----------------------- | ----------------------- |\n
            | ID             | Title                   | Severity                |\n
            | -------------- | ----------------------- | ----------------------- |\n
            | [C-01]         | Some finding             | Critical                |\n
            | -------------- | ----------------------- | ----------------------- |\n
            | [H-01]         | Some finding             | High                    |\n
            | -------------- | ----------------------- | ----------------------- |\n
            | [M-01]         | Some finding             | Medium                  |\n
            | -------------- | ----------------------- | ----------------------- |\n
            | [L-01]         | Some finding             | Low                     |\n
            | -------------- | ----------------------- | ----------------------- |\n
            | [QA-01]        | Some finding             | QA                      |\n
            | -------------- | ----------------------- | ----------------------- |\n
            | [GAS-01]       | Some finding             | GAS                     |\n
            | -------------- | ----------------------- | ----------------------- |\n
        """
        
        return TableUtils.create_html_table(
            [TableUtils.finding_to_table_row(finding) for finding in findings],
            FINDING_LIST_TABLE_COLUMNS,
            name
        )
    
class TableUtils(CustomTableUtils):
    """
        A wrapper class for the other table classes in this file.
    """
    
    @staticmethod
    def finding_to_table_row(finding: Finding) -> list[str]:
        """
            Converts a finding to a string table row.
        """
        
        # Example: '[C-01] Some critical finding'
        return [
            f"[{finding.id}]", # Example: [C-01]
            finding.title,
            finding.severity.cast_to_display_case()
        ]
    
    @staticmethod
    def tables() -> list[str]:
        """
            Returns a list of the table utilities.
            
            We need this method because if an enum gets created a circular dependency will occur.
        """
        
        return [
            "severity_classification_table",
            "information_table",
            "findings_count_table",
            "findings_summary_table"
        ]
            
    @staticmethod
    def fragment_to_callable() -> Dict[str, Callable]:
        """
            Returns a dictionary of the table utilities.
            
            We need this method because if an enum gets created a circular dependency will occur.
        """

        return {
            "severity_classification_table": TableUtils.create_severity_classification_table,
            "information_table": TableUtils.create_information_table,
            "findings_count_table": TableUtils.create_findings_count_table,
            "findings_summary_table": TableUtils.create_findings_summary_table
        }



from typing import Tuple

from inspect import getmembers, ismethod

from compiloor.services.parser.legend import ReportLegendUtils
from compiloor.services.typings.config import ProtocolInformationConfigDict
from compiloor.services.typings.finding import Severity
from compiloor.services.typings.utils import MAIN_REPORT_SECTIONS
from compiloor.services.parser.finding import Finding
from compiloor.services.typings.report import INFORMATION_TABLE_VARIABLES, REPORT_SECTION_HEADINGS
from compiloor.services.parser.table import TableUtils
from compiloor.services.environment.utils import FileUtils, FindingUtils
from compiloor.services.utils.config import ConfigUtils


class ReportCustomizer:
    """
        A class that assembles the core report. 
    """
    
    # Report configuration variables:
    config: ProtocolInformationConfigDict
    
    # Report template variables:
    report: str # The report template.
    stylesheet: str # The CSS stylesheet.
    
    # Finding report variables:
    total_findings_amount: int # The total amount of findings.
    finding_amounts_by_severity: dict[Severity, int] # The amount of findings per each severity.
    findings: str # The finding fragments in HTML format.
    serialized_findings: list[Finding] # The serialized findings in Finding object format.
    report_section_headings: list[str] # The report section headings.
    
    def __init__(self) -> str:
        # Gets the report configuration:
        self.config = FileUtils.read_config(json=True)

        # Gets the stylesheet and report template fragment:         
        self.report = FileUtils.read_file(self.config["template_url"], is_url=True)
        self.stylesheet = FileUtils.read_file(self.config["stylesheet_url"], is_url=True, html_tag="style")
        
        # Gets the finding fragments:
        (
            self.total_findings_amount,
            self.finding_amounts_by_severity,
            self.findings,
            self.serialized_findings,
            self.report_section_headings
        ) = FindingReportUtils.get_finding_fragments(self.config, 6) # 6 is the hardcoded index of the findings section: Will be changed in a future update.
        
        return self.assemble_report() # Assembles the report.
    
    def add_report_config_variables(self) -> None:
        for key in self.config.keys():
            # '{{config.<key>}}' syntax is very confusing when 
            self.report = self.report.replace(f"{{{{config.{key}}}}}", str(self.config[key]))

    def add_information_table_variables_to_report(self, tag: str = "b") -> None:
        """
            Bolds the information table row headings in the report.
        """
        
        for _key in INFORMATION_TABLE_VARIABLES.keys():
            self.report = self.report.replace(
                _key,
                f'<{tag}>{INFORMATION_TABLE_VARIABLES[_key]}</{tag}>'
            )

    def add_dynamic_tables_to_report(self) -> None:
        # Defines the table type -> callable arguments mapping:
        kwargs: dict[str, list[any]] = {
            "severity_classification_table": [],
            "information_table": [self.config],
            "findings_count_table": [self.finding_amounts_by_severity, self.total_findings_amount],
            "findings_summary_table": [self.serialized_findings],
        }
        
        # Maps the table type -> table creation callable:
        table_types = TableUtils.fragment_to_callable()
        
        # Adding the tables to the report:
        for name in table_types.keys():
            callable, args = table_types[name], kwargs[name]
            self.report = self.report.replace(f'{{{{{name}}}}}', callable(*args))
        
        # Adding no-wrap so that the table look doesn't break:
        # TODO: Change with a modular CSS stylesheet.
        self.report = self.report.replace("<td>[", '<td class="no-wrap-column">[')
        
    def add_findings_legend(self) -> None:
        # Adding the legend manually:
        self.report = self.report.replace(
            '{{findings_legend}}',
            # 6 is the hardcoded index of the findings section: Will be changed in a future update.
            ReportLegendUtils.create_finding_severities_legend(6, self.serialized_findings)
        )
        
    def assemble_report(self) -> str:
        # Adds the main parts of the report to the fragment:
        # I.E. The stylesheet, the findings, the total findings amount.
                
        section_to_attribute: dict[str, any] = {}
        for section in MAIN_REPORT_SECTIONS:
            section_to_attribute[str(section)] = getattr(self, section)
        
        for part in section_to_attribute.keys():
            self.report = self.report.replace(f"{{{{{part}}}}}", str(section_to_attribute[part]))
        
        # Getting all of the class's methods and calling all with the 'get_' prefix:
        for method in getmembers(self, predicate=lambda x: ismethod(x)):

            if not method[0].startswith("add_"): continue
            # Calling the method:
            method[1]()
        
        self.add_findings_legend()
    
class FindingReportUtils:
    """
        A class containing utilities for creating the finding report.
    """
    # OLD VERSION OF THE FUNCTION:
    
    # @staticmethod
    # def get_finding_fragments(
    #     findings_section_index: str,
    #     paragraph_subheading_tag: str = "h2"
    # ) -> Tuple[int, dict[Severity, int], str, list[Finding]]:
    #     """
    #         Gets the finding fragments from the findings directory and serializes them into different formats intended for different use.
            
    #         Returns a tuple of:
    #         - The total amount of findings.
    #         - The amount of findings per severity.
    #         - The finding fragments.
    #         - The finding summary table.
    #         - The serialized findings.
    #     """
        
    #     if type(findings_section_index) != str: findings_section_index = str(findings_section_index)
        
    #     (
    #         total_findings_amount, finding_amounts_by_severity, findings
    #     ) = FindingUtils.get_finding_fragments()
    
    #     finding_markdowns: list[str] = []
    #     _finding_fragments: list[str] = []
        
    #     __finding_fragments = findings.copy()
        
    #     for severity in __finding_fragments:
    #         if not __finding_fragments[severity]: del findings[severity]
            
    #     report_section_headings = REPORT_SECTION_HEADINGS.copy()
            
    #     for severity in findings:
            
    #         __finding_fragments = findings[severity]
    #         __finding_fragments.reverse()
    #         _finding_fragments += __finding_fragments
        
    #     _finding_fragments.reverse()
        
    #     serialized_findings: list[Finding] = [Finding(finding) for finding in _finding_fragments]
        
    #     severity_index: int = 0
    #     current_severity: Severity = None

    #     serialized_full_findings_by_severity: dict[Severity, list[Finding]] = {}
        
    #     for severity in Severity:
    #         if not severity in findings: continue
    #         serialized_full_findings_by_severity[severity] = [
    #             Finding(finding).full_fragment for finding in findings[severity]
    #         ]

    #     for finding in serialized_findings:
            
    #         severity_display = finding.severity.cast_to_display_case()
            
    #         report_section_headings.append(f"[{finding.id}] {finding.title}")
    #         if finding.severity == current_severity: continue
            
    #         current_severity = finding.severity
    #         severity_index += 1
            
    #         report_section_headings.append(f"{findings_section_index}.{severity_index}. {severity_display} Findings")

    #         serialized_full_findings_by_severity[finding.severity].reverse()
    #         finding_html_fragments = "\n".join(serialized_full_findings_by_severity[finding.severity])
    #         finding_markdowns.append(
    #             f'''<div id="section-{findings_section_index}-{severity_index}" class="page-break-after">
    #                     <{paragraph_subheading_tag} class="paragraph-subheading">
    #                         {findings_section_index}.{severity_index}. {severity_display} Findings
    #                     </{paragraph_subheading_tag}>
    #                     <div class="findings">
    #                         {finding_html_fragments}
    #                     </div>
    #                 </div>
    #             '''
    #         )        
            
    #         finding_markdowns_len = len(finding_markdowns) - 1
            
    #         finding_markdowns[finding_markdowns_len] = finding_markdowns[finding_markdowns_len].replace(
    #             f'[[{finding.severity.value}_severity_index]]',
    #             str(severity_index)
    #         )

    #     _finding_fragments.reverse()
                
    #     return (
    #         total_findings_amount, finding_amounts_by_severity, "\n".join(finding_markdowns),
    #         serialized_findings, report_section_headings
    #     )


    # NEW VERSION OF THE FUNCTION:
    @staticmethod
    def get_finding_fragments(config: ProtocolInformationConfigDict, findings_section_index: str, paragraph_subheading_tag: str = "h2") -> Tuple[int, dict[Severity, int], str, list[Finding]]:
        total_amount, amounts_by_severity, findings = FindingUtils.get_finding_fragments()

        # Remove empty severity levels
        _findings = {}

        for severity, fragments in findings.items():
            if fragments:
                _findings[severity] = fragments

        # Copy the non-empty severity levels
        findings = _findings

        markdowns, _finding_fragments = [], []

        # Copying the static headings present in the report template:
        section_headings = REPORT_SECTION_HEADINGS.copy()

        # Sort severity levels from high to low
        for severity in reversed(list(findings.keys())):
            _finding_fragments.extend(findings[severity])

        serialized = [Finding(finding) for finding in _finding_fragments]

        full_by_severity = {}
        current_severity, severity_index = None, 0

        # Iterate through severity levels from high to low
        for severity in reversed(list(Severity)):
            if not severity in findings: continue # Skip empty severity levels
            
            full_by_severity[severity] = [Finding(finding).full_fragment for finding in findings[severity]]

        for finding in serialized:
            severity_display = finding.severity.cast_to_display_case()
            section_headings.append(f"[{finding.id}] {finding.title}")

            if finding.severity == current_severity: continue # Skip if the severity level is the same as the previous one

            severity_index += 1
            current_severity = finding.severity

            # Adding the severity level heading to the section headings:
            section_headings.append(f"{findings_section_index}.{severity_index}. {severity_display} Findings")

            # Get the HTML fragments in the correct order for the current severity:
            html_fragments = "\n".join(full_by_severity[finding.severity])

            # TODO: Can this be done in a more elegant way?
            markdowns.append(
                f'''
                    <div id="section-{findings_section_index}-{severity_index}" class="page-break-after">
                        <{paragraph_subheading_tag} class="paragraph-subheading">
                            {findings_section_index}.{severity_index}. {severity_display} Findings
                        </{paragraph_subheading_tag}>
                        <div class="findings">
                            {html_fragments}
                        </div>
                    </div>
                '''
            )

            # Adding the link to the heading:
            markdowns[-1] = markdowns[-1].replace(f'[[{finding.severity.value}_severity_index]]', str(severity_index))

        section_headings[0] = section_headings[0] + " " + config["author"]

        return total_amount, amounts_by_severity, "\n".join(markdowns), serialized, section_headings


from mistune import create_markdown, escape, HTMLRenderer

from pygments import highlight
from pygments.formatters.html import HtmlFormatter
from pygments.lexers import get_lexer_by_name
from pygments.token import Error
from pygments.styles.default import DefaultStyle

from compiloor.services.logger.logger import Logger
from compiloor.services.typings.finding import Severity, SeverityFolderIndex


# This class if here to avoid circular imports:
class DefaultStyleExtended(DefaultStyle):
    """
        A class that extends the default Pygments style.
        Removes the red color from the Error token.
    """
    
    styles = DefaultStyle.styles.copy()
    styles.update({ Error: '#000' })

class Finding:
    """
        A class that represents a serialized finding.
    """
    
    id: str # The ID of the finding. Example: "[H-01]".
    id_num: int # The number of the finding in the severity it belongs to.
    
    title: str
    severity: Severity
    
    fragment: str
    
    # The full HTML fragment of the finding. It includes rendered code blocks. 
    # This is the fragment that will be inserted into the report.
    full_fragment: str 
    
    def __init__(self, fragment: str) -> None:
        self.fragment = fragment
        first_row = fragment.split("\n")[0]
        
        if not "[" in first_row or not "]" in first_row:
            Logger.error(f"Invalid finding fragment: {first_row}")
            exit(1)
        
        fragments: list[str] = first_row.split("]")
        
        if fragments[1] == "": fragments[1] = "-"
        
        self.id = fragments[0].split("[")[1].strip()
        self.id_num = int(self.id.split("-")[1].strip())
        
        self.title = fragments[1].strip()
        self.severity = SeverityFolderIndex.cast_to_severity(SeverityFolderIndex(self.id.split("-")[0].strip().upper()))

        # TODO: Abstract away the CSS classes into a modular system.
        self.full_fragment = f'''
            <div id="section-6-[[{self.severity.value}_severity_index]]-{self.id_num}" class="finding">
                {HighlightRenderer.create_markdown()(fragment)}
            </div>
        '''
                    
    def __str__(self) -> str:
        """
        Returns the finding's raw string fragment.
        """
        
        return self.fragment
    
class HighlightRenderer(HTMLRenderer):
    """
        A custom renderer that highlights code blocks.
    """
    
    def block_code(self, code, info=None):
        """
            Renders a code block with syntax highlighting.
        """
        
        if not info:
            # Wrapping it in a code block:
            return '<pre><code>' + escape(code) + '</code></pre>'
        
        lexer = get_lexer_by_name(info, stripall=True)
        formatter = HtmlFormatter(style=DefaultStyleExtended, full=True)
        
        highlighted_code: str = highlight(code, lexer, formatter)
        
        # TODO: Abstract away the CSS classes into a modular system.
        return f'<div class="code-border">{highlighted_code}</div>'

    @staticmethod
    def create_markdown(**kwargs):
        """
            Creates a Markdown parser with syntax highlighting.
        """
        
        return create_markdown(renderer=HighlightRenderer(), **kwargs)


TABLE_HTML_OUTPUT_FORMAT_SETTING = "html"

PAGE_NUMBER_FONT_SIZE = 13

PRIMARY_COLOR_IN_PERCENTAGES = (0.129, 0.145, 0.161)

REPORT_EXTENSION = ".pdf"

MAIN_REPORT_SECTIONS = [
    "findings",
    "total_findings_amount",
    "stylesheet",
]


from compiloor.services.parser.finding import Finding
from compiloor.services.typings.finding import Severity


class ReportLegendUtils:
    """
        A class that contains utility methods for creating the report legend.
    """
    
    @staticmethod
    def create_finding_severities_legend(findings_section_index: str, findings: list[Finding]) -> str:
        # The bellow code can only work with strings so we are casting the findings_section_index to a string here:
        if not isinstance(findings_section_index, str):
            findings_section_index = str(findings_section_index) 

        fragments = []
        current_severity, current_severity_index, current_severity_finding_index = None, 0, 0

        for finding in findings:
            if finding.severity != current_severity:
                
                current_severity = finding.severity
                current_severity_index += 1
                current_severity_finding_index = 0
                section_index = f'[{findings_section_index}.{current_severity_index}]_page'

                if current_severity_index != 1: fragments.append("<div>")

                fragments.extend([
                    f'''
                        <div class="sub-paragraph">
                            <a href="#section-{findings_section_index}-{current_severity_index}">
                                <div class="section-wrapper">
                                    <p>
                                        {findings_section_index}.{current_severity_index}. {current_severity.cast_to_display_case()} Findings
                                    </p>
                                    {{{{{section_index}}}}}
                                </div>
                            </a>
                        </div>
                    '''
                    # '</div>'
                ])

            current_severity_finding_index += 1
            section_index = f'[{finding.id}]_page'
            fragments.append(
                f'''
                <div class="sub-sub-paragraph">
                    <a href="#section-{findings_section_index}-{current_severity_index}-{current_severity_finding_index}">
                        <div class="section-wrapper">
                            <p>[{finding.id}] {finding.title}</p>
                            {{{{{section_index}}}}}
                        </div>
                    </a>
                </div>
                '''
            )

        return "".join(fragments)
    
    # THE OLD FUNCTION:
    # KEEPING IT HERE FOR REFERENCE.
    # @staticmethod
    # def create_finding_severities_legend(findings_section_index: str, findings: list[Finding]) -> str:
    #     """
    #         Creates a legend for the report's findings.
    #     """
    #     if type(findings_section_index) != str: findings_section_index = str(findings_section_index)
        
    #     current_severity: Severity = None
    #     current_severity_index: int = 0
    #     current_severity_finding_index: int = 0
        
    #     fragments: list[str] = []
        
    #     for finding in findings:
            
    #         if finding.severity != current_severity:
            
    #             current_severity = finding.severity
    #             current_severity_index += 1
    #             current_severity_finding_index = 0
    #             section_index = f'section_{findings_section_index}.{current_severity_index}_page'
                
    #             if current_severity_index != 1: fragments.append("</div>")
                
    #             fragments.append(
    #                 f'''
    #                 <a href="#section-{findings_section_index}-{current_severity_index}">
    #                     <div class="section-wrapper">
    #                         <h3>
    #                             {findings_section_index}.{current_severity_index}. {current_severity.cast_to_display_case()} Findings
    #                         </h3>
    #                         <p>{{{{{section_index}}}}}</p>
    #                     </div>
    #                 </a>
    #                 '''
    #             )
                
    #             fragments.append('<div class="sub-sub-paragraph">')
            
    #         current_severity_finding_index += 1
            
    #         section_index = f'section_[{finding.id}]_page'
            
    #         fragments.append(
    #             f'''
    #                 <a href="#section-{findings_section_index}-{current_severity_index}-{current_severity_finding_index}">
    #                     <div class="section-wrapper">
    #                         <h3>
    #                             [{finding.id}] {finding.title}
    #                         </h3>
    #                         <p>{{{{{section_index}}}}}</p>
    #                     </div>
    #                 </a>         
    #             '''
    #         )
            
    #     return "".join(fragments)


from json import loads

from os import listdir
from os.path import abspath

from typing import Tuple

from requests import get

from compiloor.services.environment.constants import (
    CONFIG_DIRECTORY, FINDINGS_DIRECTORY, REPORTS_DIRECTORY
)
from compiloor.services.typings.config import ProtocolInformationConfigDict
from compiloor.services.typings.finding import Severity
from compiloor.services.typings.utils import REPORT_EXTENSION

class FileUtils:
    """
        Utility class for file operations.
    """
    
    @staticmethod
    def read_config(json: bool = False) -> str | ProtocolInformationConfigDict:
        """
            Returns the serialized contents of the config file.
        """
        
        _config = open(abspath(CONFIG_DIRECTORY), "r").read()
        return loads(_config) if json else _config
    
    @staticmethod
    def read_finding(severity: Severity, index: int) -> str:
        """
            Returns the contents of the finding with the given severity and index.
        """
        
        return open(abspath(FINDINGS_DIRECTORY) + f"/[{severity.cast_to_folder_sig().value}-{FileUtils.get_fs_sig_index(index)}].md", "r").read()
    
    @staticmethod
    def read_file(name: str, is_url: bool = False, html_tag: str | None = None) -> str:
        """
            Returns the contents of the file with the given name. Wrapps them in the given HTML tag if provided.
        """

        file: str = open(abspath(name), "r").read() if not is_url else get(name).content.decode("utf-8")
        
        if html_tag: file = f"<{html_tag}>{file}</{html_tag}>"
        
        return file
    
    @staticmethod
    def get_fs_sig_index(index: int) -> str:
        """
            Returns the report index in the proper signature format.
        """
        
        return "0" + str(index) if index < 10 else str(index)
    

class FindingUtils:
    """
        Utility class for findings.
    """
    
    @staticmethod
    def get_current_finding_amount(severity: Severity) -> int:
        """ 
            Returns the current amount of findings for the given severity.
        """
        return len(
            list(
                filter(
                    lambda x: x.endswith(".md") and x.startswith(f"[{severity.cast_to_folder_sig().value}"), listdir(abspath(FINDINGS_DIRECTORY))
                )
            )
        )
    
    @staticmethod
    def get_finding_fragments() -> Tuple[int, dict[Severity, int], dict[Severity, list[str]]]:
        
        """
            Returns a tuple of two dictionaries:
            - The first one contains the amount of findings per severity.
            - The second one contains the findings themselves.
        """
        
        total_findings: int = 0
        
        findings_by_severity_amount: dict[Severity, int] = {}
        findings_by_severity: dict[Severity, str] = {}
        
        for _severity in Severity:
            findings_amount_by_severity = FindingUtils.get_current_finding_amount(_severity)
            findings_by_severity_amount[_severity] = findings_amount_by_severity
            total_findings += findings_amount_by_severity
            findings_by_severity[_severity] = [
                FileUtils.read_finding(_severity, finding) for finding in range(1, findings_amount_by_severity + 1)
            ] 
            
        return [
            total_findings, findings_by_severity_amount, findings_by_severity
        ]        
    
class ReportUtils:
    """
        Class with utility functions for report generation.
    """
    
    @staticmethod
    def get_current_report_count() -> int:
        """
            Returns the current report count.
        """
        
        return len(listdir(abspath(REPORTS_DIRECTORY)))
    
    @staticmethod
    def get_report_name(index: int | None) -> str:
        """
            Returns the name of the report with the given index.
        """
        
        if not index: index = ReportUtils.get_current_report_count()
        return f'{abspath(REPORTS_DIRECTORY)}/report-{FileUtils.get_fs_sig_index(index)}{REPORT_EXTENSION}'


from bs4 import BeautifulSoup

from typing import Tuple

from os.path import abspath

from playwright.sync_api import sync_playwright

from compiloor.services.environment.utils import FileUtils, ReportUtils
from compiloor.services.typings.utils import REPORT_EXTENSION
from compiloor.services.environment.constants import REPORTS_DIRECTORY


class ChromiumUtils:
    """
        A class containing utilities for creating PDF documents from HTML fragments using the playwright chromium driver.
    """
    
    @staticmethod
    def create_chromium_document(fragment: str, dir: str = abspath(REPORTS_DIRECTORY)) -> Tuple[int, str]:
        """
            Creates a PDF document from the given HTML fragment and saves it in the given directory.
        """
        fragment = BeautifulSoup(fragment, "html.parser")

        # Using a context manager just because the documentation says so:
        with sync_playwright() as playwright:
            browser = playwright.chromium.launch(headless=True)
            document = browser.new_page()
            document.set_content(str(fragment), wait_until="networkidle")
            report_index = ReportUtils.get_current_report_count() + 1 
            
            dir = f'{dir}/report-{FileUtils.get_fs_sig_index(report_index)}{REPORT_EXTENSION}'
            
            document.pdf(path=dir)
            # # Keeping the html digest for debugging purposes:
            # open("report.html", "w").write(document.content())
            browser.close()
        return report_index, dir


from typing import TypedDict

class ProtocolInformationConfigDict(TypedDict):
    title: str
    author: str
    date: str
    protocol_name: str
    repository: str
    type: str
    commit: str
    sloc: str
    about_content: str
    disclaimer_content: str
    template_url: str
    stylesheet_url: str
    cover_img_url: str
    